(* JTG main loop.
   Copyright (C) 2003 Jean Goubault-Larrecq and LSV, CNRS UMR 8643 & ENS Cachan.

   This file is part of jtg.

   jtg is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   jtg is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with jtg; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*)

val bad_sequences = ref (nil : (string list * num * num * int ref) list);
    (* for every file describing a bad sequence (or even just one event),
     - list of events;
     - rate at which new bad sequences are initiated, or NONE if just once;
     - rate at which events from the bad sequence are inserted into the good sequence;
     - max number of times where this sequence can be fired again;
     *)
val maxoutput = ref max_int;

exception QuitEvt of int;

fun ioerr (s, n) =
    (#put stderr s;
     #put stderr ": ";
     #put stderr (iomsg n);
     #put stderr "\n";
     #flush stderr ();
     raise QuitEvt 2);

fun openfile s =
    infile s handle IO n => ioerr (s, n);

fun readfile s =
    let val f as |[getline, ...]| = openfile s
	fun getlines acc =
	    let val l = getline ()
	    in
		if l=""
		    then acc
		else getlines (l::acc)
	    end
    in
	getlines nil
    end;

val |[matches = isint, ...]| = regexp "^[0-9]+$";

fun getint s = (* where s matches the above regexp. *)
    let val i = intofstring s
    in
	i
    end;

val |[matches = isfloat, ...]| = regexp "^[0-9]+(\\.[0-9]*)?([eEfFgG][~-]?[0-9]+)?%?$";

fun getrate s = (* where s matches the above regexp. *)
    let val x0 = numofstring s
	val x = if ordof (s, size s-1)=ord "%"
		    then 0.01 #* x0
		else x0
    in
	if x #> 0.0 andalso x #<= 1.0
	    then x
	else (#put stderr s;
	      #put stderr " should be a number between 0 and 1, or a percentage between 0% and 100%\n";
	      #flush stderr ();
	      raise QuitEvt 2)
    end;

fun usage () =
    (#put stderr "Usage: jtgmix flags* [bad-sequence-file init-rate event-rate max-triggered]* [event-file]?.\n\
     \  Version 1.0, Copyright (C) Jean Goubault-Larrecq;\n\
     \     see file COPYRIGHT.\n\
     \  jtg comes with ABSOLUTELY NO WARRANTY; see file COPYING, sections 11, 12.\n\
     \  This is free software, and you are welcome to redistribute it\n\
     \  under certain conditions; see TERMS AND CONDITIONS in file COPYING.\n\
     \  An event-file argument of '-' stands for standard input.  Flags may be:\n\
     \  -max <n>: outputs at most n lines.\n\
     \  -h: prints this help.\n";
     #flush stderr ());

fun infile_or_stdin "-" = stdin
  | infile_or_stdin name =
    let val |[get, getline, ...]| = openfile name
    in
	|[get=get, getline=getline]|
    end;

exception ExitDoJtgEvt;

fun do_jtg (opt_f, bads) =
    let val current = ref (nil : (string list * num) list)
	fun put s =
	    (dec maxoutput;
	     if !maxoutput<0
		 then raise ExitDoJtgEvt
	     else #put stdout s)
	fun new_sequences () =
	    let val alive = ref false
	    in
		iterate
		  (alive := true;
		   if random () #< initrate
		       then (dec r;
			     current := (events, evtrate) :: !current)
		   else ())
		| (events, initrate, evtrate, r as ref n) in list bads
		    such that n>0 andalso not (null events)
		end;
		if !alive
		    then ()
		else raise ExitDoJtgEvt
	    end
	fun advance_sequences () =
	    current := [if random () #< evtrate
			    then (put evt1;
				  (others, evtrate))
			else e
		       | e as (evt1::others, evtrate) in list !current]
	val l = ref ""
    in
	(case opt_f of
	     SOME |[getline, ...]| =>
	     while (l := getline (); !l <> "") do
		 (new_sequences (); advance_sequences ();
		  put (!l))
	   | _ =>
	     while true do
		 (new_sequences (); advance_sequences ()))
	     handle ExitDoJtgEvt => while not (null (!current)) do advance_sequences ();
	     #flush stdout ()
    end;

fun do_args ("-max"::s::rest) =
    (maxoutput := getint s;
     do_args rest)
  | do_args ("-h" :: rest) =
    (usage (); raise QuitEvt 0)
  | do_args (bad::initrate::evtrate::maxtrig::rest) =
    (bad_sequences := (rev (readfile bad),
		       getrate initrate,
		       getrate evtrate,
		       ref (getint maxtrig)) :: !bad_sequences;
     do_args rest)
  | do_args [goodfile] = do_jtg (SOME (infile_or_stdin goodfile), !bad_sequences)
  | do_args nil =
    if null (!bad_sequences)
	then (usage (); raise QuitEvt 0)
    else do_jtg (NONE, !bad_sequences)
  | do_args _ = (usage (); raise QuitEvt 2);

fun main () =
    do_args (args ()) handle QuitEvt n => quit n;
